"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EquityDto = exports.PerpEquity = exports.TokenEquity = exports.Equity = exports.MangoAccountData = exports.HealthType = exports.PerpPositionDto = exports.PerpPositions = exports.Serum3PositionDto = exports.Serum3Orders = exports.TokenPositionDto = exports.TokenPosition = exports.MangoAccount = void 0;
const bytes_1 = require("@project-serum/anchor/dist/cjs/utils/bytes");
const utils_1 = require("../utils");
const bank_1 = require("./bank");
const healthCache_1 = require("./healthCache");
const I80F48_1 = require("./I80F48");
class MangoAccount {
    constructor(publicKey, group, owner, name, delegate, beingLiquidated, isBankrupt, accountNum, bump, netDeposits, netSettled, headerVersion, tokens, serum3, perps, perpOpenOrders, accountData) {
        this.publicKey = publicKey;
        this.group = group;
        this.owner = owner;
        this.delegate = delegate;
        this.accountNum = accountNum;
        this.accountData = accountData;
        this.name = bytes_1.utf8.decode(new Uint8Array(name)).split('\x00')[0];
        this.tokens = tokens.map((dto) => TokenPosition.from(dto));
        this.serum3 = serum3.map((dto) => Serum3Orders.from(dto));
        this.perps = perps.map((dto) => PerpPositions.from(dto));
    }
    static from(publicKey, obj) {
        return new MangoAccount(publicKey, obj.group, obj.owner, obj.name, obj.delegate, obj.beingLiquidated, obj.isBankrupt, obj.accountNum, obj.bump, obj.netDeposits, obj.netSettled, obj.headerVersion, obj.tokens, obj.serum3, obj.perps, obj.perpOpenOrders, {});
    }
    async reload(client, group) {
        Object.assign(this, await client.getMangoAccount(this));
        await this.reloadAccountData(client, group);
    }
    async reloadAccountData(client, group) {
        this.accountData = await client.computeAccountData(group, this);
    }
    findToken(tokenIndex) {
        return this.tokens.find((ta) => ta.tokenIndex == tokenIndex);
    }
    findSerum3Account(marketIndex) {
        return this.serum3.find((sa) => sa.marketIndex == marketIndex);
    }
    getNative(bank) {
        const ta = this.findToken(bank.tokenIndex);
        return ta ? ta.native(bank) : I80F48_1.ZERO_I80F48;
    }
    static getEquivalentNativeUsdcPosition(sourceBank, nativeTokenPosition) {
        return nativeTokenPosition
            ? nativeTokenPosition
                .native(sourceBank)
                .mul(I80F48_1.I80F48.fromNumber(Math.pow(10, bank_1.QUOTE_DECIMALS)))
                .div(I80F48_1.I80F48.fromNumber(Math.pow(10, sourceBank.mintDecimals)))
                .mul(sourceBank.price)
            : I80F48_1.ZERO_I80F48;
    }
    static getEquivalentNativeTokenPosition(targetBank, nativeUsdcPosition) {
        return nativeUsdcPosition
            .div(targetBank.price)
            .div(I80F48_1.I80F48.fromNumber(Math.pow(10, bank_1.QUOTE_DECIMALS)))
            .mul(I80F48_1.I80F48.fromNumber(Math.pow(10, targetBank.mintDecimals)));
    }
    getNativeDeposits(bank) {
        const native = this.getNative(bank);
        return native.gte(I80F48_1.ZERO_I80F48) ? native : I80F48_1.ZERO_I80F48;
    }
    getNativeBorrows(bank) {
        const native = this.getNative(bank);
        return native.lte(I80F48_1.ZERO_I80F48) ? native : I80F48_1.ZERO_I80F48;
    }
    getUi(bank) {
        const ta = this.findToken(bank.tokenIndex);
        return ta ? ta.ui(bank) : 0;
    }
    deposits(bank) {
        const ta = this.findToken(bank.tokenIndex);
        return ta ? ta.uiDeposits(bank) : 0;
    }
    borrows(bank) {
        const ta = this.findToken(bank.tokenIndex);
        return ta ? ta.uiBorrows(bank) : 0;
    }
    getHealth(healthType) {
        return healthType == HealthType.init
            ? this.accountData.initHealth
            : this.accountData.maintHealth;
    }
    /**
     * TODO: this is incorrect, getAssetsVal and getLiabsVal are in equity, and not in given health type.
     * Wait for dev to be deployed to mainnet, and then we can adapt this.
     */
    getHealthRatio(healthType) {
        const assets = this.getAssetsVal();
        const liabs = this.getLiabsVal();
        return liabs.gt(I80F48_1.ZERO_I80F48)
            ? assets.div(liabs).sub(I80F48_1.ONE_I80F48).mul(I80F48_1.I80F48.fromNumber(100))
            : I80F48_1.I80F48.fromNumber(100);
    }
    /**
     * Sum of all the assets i.e. token deposits, borrows, total assets in spot open orders, (perps positions is todo) in terms of quote value.
     */
    getEquity() {
        const equity = this.accountData.equity;
        const total_equity = equity.tokens.reduce((a, b) => a.add(b.value), I80F48_1.ZERO_I80F48);
        return total_equity;
    }
    /**
     * The amount of native quote you could withdraw against your existing assets.
     */
    getCollateralValue() {
        return this.getHealth(HealthType.init);
    }
    /**
     * Similar to getEquity, but only the sum of all positive assets.
     */
    getAssetsVal() {
        const equity = this.accountData.equity;
        const total_equity = equity.tokens.reduce((a, b) => (b.value.gt(I80F48_1.ZERO_I80F48) ? a.add(b.value) : a), I80F48_1.ZERO_I80F48);
        return total_equity;
    }
    /**
     * Similar to getEquity, but only the sum of all negative assets. Note: return value would be negative.
     */
    getLiabsVal() {
        const equity = this.accountData.equity;
        const total_equity = equity.tokens.reduce((a, b) => (b.value.lt(I80F48_1.ZERO_I80F48) ? a.add(b.value) : a), I80F48_1.ZERO_I80F48);
        return total_equity;
    }
    /**
     * The amount of given native token you can borrow, considering all existing assets as collateral except the deposits for this token.
     * Note 1: The existing native deposits need to be added to get the full amount that could be withdrawn.
     * Note 2: The group might have less native deposits than what this returns. TODO: loan origination fees
     */
    getMaxWithdrawWithBorrowForToken(group, tokenName) {
        const bank = group.banksMap.get(tokenName);
        const initHealth = this.accountData.initHealth;
        const inUsdcUnits = MangoAccount.getEquivalentNativeUsdcPosition(bank, this.findToken(bank.tokenIndex)).max(I80F48_1.ZERO_I80F48);
        const newInitHealth = initHealth.sub(inUsdcUnits.mul(bank.initAssetWeight));
        return MangoAccount.getEquivalentNativeTokenPosition(bank, newInitHealth.div(bank.initLiabWeight));
    }
    /**
     * The amount of given source native token you can swap to a target token considering all existing assets as collateral.
     * note: slippageAndFeesFactor is a normalized number, <1, e.g. a slippage of 5% and some fees which are 1%, then slippageAndFeesFactor = 0.94
     * the factor is used to compute how much target can be obtained by swapping source TODO: loan origination fees
     */
    getMaxSourceForTokenSwap(group, sourceTokenName, targetTokenName, slippageAndFeesFactor) {
        const initHealth = this.accountData.initHealth;
        const sourceBank = group.banksMap.get(sourceTokenName);
        const targetBank = group.banksMap.get(targetTokenName);
        // This is a conservative approximation of the easy case, where
        // mango account has no token positions for source and target tokens, or
        // borrows for source and deposits for target tokens before the swap.
        // Tighter estimates can be obtained by adding cases where deposits can exist for source,
        // and borrows for target. TODO: solve this by searching over a blackbox like health formula.
        // Lets solve below for s,
        // h - s * slw + t * taw = 0
        // where h is init_health, s is source amount in usdc native units, and t is target amount in usdc native units
        // where t = s * slip ( s < 1), where slip is factor for slippage and fees which is normalised e.g. for 5% slippage, slip = 0.95
        // h - s * (slw - slip * taw) = 0
        // s = h / ( slw - slip * taw )
        return initHealth
            .div(sourceBank.initLiabWeight.sub(I80F48_1.I80F48.fromNumber(slippageAndFeesFactor).mul(targetBank.initAssetWeight)))
            .div(sourceBank.price);
    }
    /**
     * Simulates new health after applying tokenChanges to the token positions. Useful to simulate health after a potential swap.
     */
    simHealthWithTokenPositionChanges(group, tokenChanges) {
        // This is a approximation of the easy case, where
        // mango account has no token positions for tokens in changes list, or
        // the change is in direction e.g. deposits for deposits, borrows for borrows, of existing token position.
        // TODO: recompute entire health using components.
        const initHealth = this.accountData.initHealth;
        for (const change of tokenChanges) {
            const bank = group.banksMap.get(change.tokenName);
            if (change.tokenAmount >= 0) {
                initHealth.add(bank.initAssetWeight
                    .mul(I80F48_1.I80F48.fromNumber(change.tokenAmount))
                    .mul(bank.price));
            }
            else {
                initHealth.sub(bank.initLiabWeight
                    .mul(I80F48_1.I80F48.fromNumber(change.tokenAmount))
                    .mul(bank.price));
            }
        }
        return initHealth;
    }
    /**
     * The remaining native quote margin available for given market.
     *
     * TODO: this is a very bad estimation atm.
     * It assumes quote asset is always USDC,
     * it assumes that there are no interaction effects,
     * it assumes that there are no existing borrows for either of the tokens in the market.
     */
    getSerum3MarketMarginAvailable(group, marketName) {
        const initHealth = this.accountData.initHealth;
        const serum3Market = group.serum3MarketsMap.get(marketName);
        const marketAssetWeight = group.findBank(serum3Market.baseTokenIndex).initAssetWeight;
        return initHealth.div(I80F48_1.ONE_I80F48.sub(marketAssetWeight));
    }
    /**
     * The remaining native quote margin available for given market.
     *
     * TODO: this is a very bad estimation atm.
     * It assumes quote asset is always USDC,
     * it assumes that there are no interaction effects,
     * it assumes that there are no existing borrows for either of the tokens in the market.
     */
    getPerpMarketMarginAvailable(group, marketName) {
        const initHealth = this.accountData.initHealth;
        const perpMarket = group.perpMarketsMap.get(marketName);
        const marketAssetWeight = perpMarket.initAssetWeight;
        return initHealth.div(I80F48_1.ONE_I80F48.sub(marketAssetWeight));
    }
    tokensActive() {
        return this.tokens.filter((token) => token.isActive());
    }
    serum3Active() {
        return this.serum3.filter((serum3) => serum3.isActive());
    }
    perpActive() {
        return this.perps.filter((perp) => perp.isActive());
    }
    toString(group) {
        let res = 'MangoAccount';
        res = res + '\n pk: ' + this.publicKey.toString();
        res = res + '\n name: ' + this.name;
        res = res + '\n delegate: ' + this.delegate;
        res =
            this.tokensActive().length > 0
                ? res +
                    '\n tokens:' +
                    JSON.stringify(this.tokensActive().map((token) => token.toString(group)), null, 4)
                : res + '';
        res =
            this.serum3Active().length > 0
                ? res + '\n serum:' + JSON.stringify(this.serum3Active(), null, 4)
                : res + '';
        res =
            this.perpActive().length > 0
                ? res + '\n perps:' + JSON.stringify(this.perpActive(), null, 4)
                : res + '';
        return res;
    }
}
exports.MangoAccount = MangoAccount;
class TokenPosition {
    constructor(indexedPosition, tokenIndex, inUseCount) {
        this.indexedPosition = indexedPosition;
        this.tokenIndex = tokenIndex;
        this.inUseCount = inUseCount;
    }
    static from(dto) {
        return new TokenPosition(I80F48_1.I80F48.from(dto.indexedPosition), dto.tokenIndex, dto.inUseCount);
    }
    isActive() {
        return this.tokenIndex !== 65535;
    }
    native(bank) {
        if (this.indexedPosition.isPos()) {
            return bank.depositIndex.mul(this.indexedPosition);
        }
        else {
            return bank.borrowIndex.mul(this.indexedPosition);
        }
    }
    ui(bank) {
        return (0, utils_1.nativeI80F48ToUi)(this.native(bank), bank.mintDecimals).toNumber();
    }
    uiDeposits(bank) {
        return (0, utils_1.nativeI80F48ToUi)(bank.depositIndex.mul(this.indexedPosition), bank.mintDecimals).toNumber();
    }
    uiBorrows(bank) {
        return (0, utils_1.nativeI80F48ToUi)(bank.borrowIndex.mul(this.indexedPosition), bank.mintDecimals).toNumber();
    }
    toString(group) {
        let extra = '';
        if (group) {
            const bank = group.findBank(this.tokenIndex);
            if (bank) {
                const native = this.native(bank);
                extra += ', native: ' + native.toNumber();
                extra += ', ui: ' + this.ui(bank);
                extra += ', tokenName: ' + bank.name;
            }
        }
        return ('tokenIndex: ' +
            this.tokenIndex +
            ', inUseCount: ' +
            this.inUseCount +
            ', indexedValue: ' +
            this.indexedPosition.toNumber() +
            extra);
    }
}
exports.TokenPosition = TokenPosition;
TokenPosition.TokenIndexUnset = 65535;
class TokenPositionDto {
    constructor(indexedPosition, tokenIndex, inUseCount, reserved) {
        this.indexedPosition = indexedPosition;
        this.tokenIndex = tokenIndex;
        this.inUseCount = inUseCount;
        this.reserved = reserved;
    }
}
exports.TokenPositionDto = TokenPositionDto;
class Serum3Orders {
    constructor(openOrders, marketIndex, baseTokenIndex, quoteTokenIndex) {
        this.openOrders = openOrders;
        this.marketIndex = marketIndex;
        this.baseTokenIndex = baseTokenIndex;
        this.quoteTokenIndex = quoteTokenIndex;
    }
    static from(dto) {
        return new Serum3Orders(dto.openOrders, dto.marketIndex, dto.baseTokenIndex, dto.quoteTokenIndex);
    }
    isActive() {
        return this.marketIndex !== Serum3Orders.Serum3MarketIndexUnset;
    }
}
exports.Serum3Orders = Serum3Orders;
Serum3Orders.Serum3MarketIndexUnset = 65535;
class Serum3PositionDto {
    constructor(openOrders, marketIndex, baseTokenIndex, quoteTokenIndex, reserved) {
        this.openOrders = openOrders;
        this.marketIndex = marketIndex;
        this.baseTokenIndex = baseTokenIndex;
        this.quoteTokenIndex = quoteTokenIndex;
        this.reserved = reserved;
    }
}
exports.Serum3PositionDto = Serum3PositionDto;
class PerpPositions {
    constructor(marketIndex, basePositionLots, quotePositionNative, bidsBaseLots, asksBaseLots, takerBaseLots, takerQuoteLots) {
        this.marketIndex = marketIndex;
        this.basePositionLots = basePositionLots;
        this.quotePositionNative = quotePositionNative;
        this.bidsBaseLots = bidsBaseLots;
        this.asksBaseLots = asksBaseLots;
        this.takerBaseLots = takerBaseLots;
        this.takerQuoteLots = takerQuoteLots;
    }
    static from(dto) {
        return new PerpPositions(dto.marketIndex, dto.basePositionLots.toNumber(), dto.quotePositionNative.val.toNumber(), dto.bidsBaseLots.toNumber(), dto.asksBaseLots.toNumber(), dto.takerBaseLots.toNumber(), dto.takerQuoteLots.toNumber());
    }
    isActive() {
        return this.marketIndex != PerpPositions.PerpMarketIndexUnset;
    }
}
exports.PerpPositions = PerpPositions;
PerpPositions.PerpMarketIndexUnset = 65535;
class PerpPositionDto {
    constructor(marketIndex, reserved, basePositionLots, quotePositionNative, bidsBaseLots, asksBaseLots, takerBaseLots, takerQuoteLots) {
        this.marketIndex = marketIndex;
        this.reserved = reserved;
        this.basePositionLots = basePositionLots;
        this.quotePositionNative = quotePositionNative;
        this.bidsBaseLots = bidsBaseLots;
        this.asksBaseLots = asksBaseLots;
        this.takerBaseLots = takerBaseLots;
        this.takerQuoteLots = takerQuoteLots;
    }
}
exports.PerpPositionDto = PerpPositionDto;
class HealthType {
}
exports.HealthType = HealthType;
HealthType.maint = { maint: {} };
HealthType.init = { init: {} };
class MangoAccountData {
    constructor(healthCache, initHealth, maintHealth, equity) {
        this.healthCache = healthCache;
        this.initHealth = initHealth;
        this.maintHealth = maintHealth;
        this.equity = equity;
    }
    static from(event) {
        return new MangoAccountData(new healthCache_1.HealthCache(event.healthCache), I80F48_1.I80F48.from(event.initHealth), I80F48_1.I80F48.from(event.maintHealth), Equity.from(event.equity));
    }
}
exports.MangoAccountData = MangoAccountData;
class Equity {
    constructor(tokens, perps) {
        this.tokens = tokens;
        this.perps = perps;
    }
    static from(dto) {
        return new Equity(dto.tokens.map((token) => new TokenEquity(token.tokenIndex, I80F48_1.I80F48.from(token.value))), dto.perps.map((perpAccount) => new PerpEquity(perpAccount.perpMarketIndex, I80F48_1.I80F48.from(perpAccount.value))));
    }
}
exports.Equity = Equity;
class TokenEquity {
    constructor(tokenIndex, value) {
        this.tokenIndex = tokenIndex;
        this.value = value;
    }
}
exports.TokenEquity = TokenEquity;
class PerpEquity {
    constructor(perpMarketIndex, value) {
        this.perpMarketIndex = perpMarketIndex;
        this.value = value;
    }
}
exports.PerpEquity = PerpEquity;
class EquityDto {
}
exports.EquityDto = EquityDto;
//# sourceMappingURL=mangoAccount.js.map