"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MangoClient = void 0;
const anchor_1 = require("@project-serum/anchor");
const serum_1 = require("@project-serum/serum");
const token_instructions_1 = require("@project-serum/serum/lib/token-instructions");
const spl_token_1 = require("@solana/spl-token");
const web3_js_1 = require("@solana/web3.js");
const bs58_1 = __importDefault(require("bs58"));
const bank_1 = require("./accounts/bank");
const group_1 = require("./accounts/group");
const I80F48_1 = require("./accounts/I80F48");
const mangoAccount_1 = require("./accounts/mangoAccount");
const oracle_1 = require("./accounts/oracle");
const perp_1 = require("./accounts/perp");
const serum3_1 = require("./accounts/serum3");
const constants_1 = require("./constants");
const ids_1 = require("./ids");
const mango_v4_1 = require("./mango_v4");
const utils_1 = require("./utils");
const anchor_2 = require("./utils/anchor");
var AccountRetriever;
(function (AccountRetriever) {
    AccountRetriever[AccountRetriever["Scanning"] = 0] = "Scanning";
    AccountRetriever[AccountRetriever["Fixed"] = 1] = "Fixed";
})(AccountRetriever || (AccountRetriever = {}));
// TODO: replace ui values with native as input wherever possible
// TODO: replace token/market names with token or market indices
class MangoClient {
    constructor(program, programId, cluster, groupName) {
        this.program = program;
        this.programId = programId;
        this.cluster = cluster;
        this.groupName = groupName;
        // TODO: evil side effect, but limited backtraces are a nightmare
        Error.stackTraceLimit = 1000;
    }
    /// public
    // Group
    async groupCreate(groupNum, testing, version, insuranceMintPk) {
        const adminPk = this.program.provider.wallet.publicKey;
        return await this.program.methods
            .groupCreate(groupNum, testing ? 1 : 0, version)
            .accounts({
            creator: adminPk,
            payer: adminPk,
            insuranceMint: insuranceMintPk,
        })
            .rpc();
    }
    async groupEdit(group, newAdmin, newFastListingAdmin) {
        return await this.program.methods
            .groupEdit(newAdmin, newFastListingAdmin)
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
        })
            .rpc();
    }
    async groupClose(group) {
        const adminPk = this.program.provider.wallet.publicKey;
        return await this.program.methods
            .groupClose()
            .accounts({
            group: group.publicKey,
            insuranceVault: group.insuranceVault,
            admin: adminPk,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .rpc();
    }
    async getGroup(groupPk) {
        const groupAccount = await this.program.account.group.fetch(groupPk);
        const group = group_1.Group.from(groupPk, groupAccount);
        await group.reloadAll(this);
        return group;
    }
    async getGroupForCreator(creatorPk, groupNum) {
        const filters = [
            {
                memcmp: {
                    bytes: creatorPk.toBase58(),
                    offset: 8,
                },
            },
        ];
        if (groupNum) {
            const bbuf = Buffer.alloc(4);
            bbuf.writeUInt32LE(groupNum);
            filters.push({
                memcmp: {
                    bytes: bs58_1.default.encode(bbuf),
                    offset: 40,
                },
            });
        }
        const groups = (await this.program.account.group.all(filters)).map((tuple) => group_1.Group.from(tuple.publicKey, tuple.account));
        await groups[0].reloadAll(this);
        return groups[0];
    }
    // Tokens/Banks
    async tokenRegister(group, mintPk, oraclePk, oracleConfFilter, tokenIndex, name, adjustmentFactor, util0, rate0, util1, rate1, maxRate, loanFeeRate, loanOriginationFeeRate, maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee) {
        return await this.program.methods
            .tokenRegister(tokenIndex, name, {
            confFilter: {
                val: I80F48_1.I80F48.fromNumber(oracleConfFilter).getData(),
            },
        }, // future: nested custom types dont typecheck, fix if possible?
        { adjustmentFactor, util0, rate0, util1, rate1, maxRate }, loanFeeRate, loanOriginationFeeRate, maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee)
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            mint: mintPk,
            oracle: oraclePk,
            payer: this.program.provider.wallet.publicKey,
            rent: web3_js_1.SYSVAR_RENT_PUBKEY,
        })
            .rpc();
    }
    async tokenRegisterTrustless(group, mintPk, oraclePk, tokenIndex, name) {
        return await this.program.methods
            .tokenRegisterTrustless(tokenIndex, name)
            .accounts({
            group: group.publicKey,
            fastListingAdmin: this.program.provider.wallet
                .publicKey,
            mint: mintPk,
            oracle: oraclePk,
            payer: this.program.provider.wallet.publicKey,
            rent: web3_js_1.SYSVAR_RENT_PUBKEY,
        })
            .rpc();
    }
    async tokenEdit(group, tokenName, oracle, oracleConfFilter, groupInsuranceFund, adjustmentFactor, util0, rate0, util1, rate1, maxRate, loanFeeRate, loanOriginationFeeRate, maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee) {
        const bank = group.banksMap.get(tokenName);
        const mintInfo = group.mintInfosMap.get(bank.tokenIndex);
        return await this.program.methods
            .tokenEdit(new anchor_1.BN(0), oracle, {
            confFilter: {
                val: I80F48_1.I80F48.fromNumber(oracleConfFilter).getData(),
            },
        }, // future: nested custom types dont typecheck, fix if possible?
        groupInsuranceFund !== null && groupInsuranceFund !== void 0 ? groupInsuranceFund : null, { adjustmentFactor, util0, rate0, util1, rate1, maxRate }, loanFeeRate, loanOriginationFeeRate, maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee)
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            mintInfo: mintInfo.publicKey,
        })
            .remainingAccounts([
            {
                pubkey: bank.publicKey,
                isWritable: true,
                isSigner: false,
            },
        ])
            .rpc({ skipPreflight: true });
    }
    async tokenDeregister(group, tokenName) {
        var _a;
        const bank = group.banksMap.get(tokenName);
        const adminPk = this.program.provider.wallet.publicKey;
        const dustVaultPk = await (0, utils_1.getAssociatedTokenAddress)(bank.mint, adminPk);
        const ai = await this.program.provider.connection.getAccountInfo(dustVaultPk);
        if (!ai) {
            const tx = new web3_js_1.Transaction();
            tx.add(spl_token_1.Token.createAssociatedTokenAccountInstruction(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, bank.mint, dustVaultPk, adminPk, adminPk));
            await this.program.provider.sendAndConfirm(tx);
        }
        return await this.program.methods
            .tokenDeregister(bank.tokenIndex)
            .accounts({
            group: group.publicKey,
            admin: adminPk,
            mintInfo: (_a = group.mintInfosMap.get(bank.tokenIndex)) === null || _a === void 0 ? void 0 : _a.publicKey,
            dustVault: dustVaultPk,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .remainingAccounts([bank.publicKey, bank.vault].map((pk) => ({ pubkey: pk, isWritable: true, isSigner: false })))
            .rpc();
    }
    async getBanksForGroup(group) {
        return (await this.program.account.bank.all([
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ])).map((tuple) => bank_1.Bank.from(tuple.publicKey, tuple.account));
    }
    async getMintInfosForGroup(group) {
        return (await this.program.account.mintInfo.all([
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ])).map((tuple) => {
            return bank_1.MintInfo.from(tuple.publicKey, tuple.account);
        });
    }
    async getMintInfoForTokenIndex(group, tokenIndex) {
        const tokenIndexBuf = Buffer.alloc(2);
        tokenIndexBuf.writeUInt16LE(tokenIndex);
        return (await this.program.account.mintInfo.all([
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
            {
                memcmp: {
                    bytes: bs58_1.default.encode(tokenIndexBuf),
                    offset: 40,
                },
            },
        ])).map((tuple) => {
            return bank_1.MintInfo.from(tuple.publicKey, tuple.account);
        });
    }
    // Stub Oracle
    async stubOracleCreate(group, mintPk, price) {
        return await this.program.methods
            .stubOracleCreate({ val: I80F48_1.I80F48.fromNumber(price).getData() })
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            mint: mintPk,
            payer: this.program.provider.wallet.publicKey,
        })
            .rpc();
    }
    async stubOracleClose(group, oracle) {
        return await this.program.methods
            .stubOracleClose()
            .accounts({
            group: group.publicKey,
            oracle: oracle,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .rpc();
    }
    async stubOracleSet(group, oraclePk, price) {
        return await this.program.methods
            .stubOracleSet({ val: I80F48_1.I80F48.fromNumber(price).getData() })
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            oracle: oraclePk,
            payer: this.program.provider.wallet.publicKey,
        })
            .rpc();
    }
    async getStubOracle(group, mintPk) {
        const filters = [
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ];
        if (mintPk) {
            filters.push({
                memcmp: {
                    bytes: mintPk.toBase58(),
                    offset: 40,
                },
            });
        }
        return (await this.program.account.stubOracle.all(filters)).map((pa) => oracle_1.StubOracle.from(pa.publicKey, pa.account));
    }
    // MangoAccount
    async getOrCreateMangoAccount(group, ownerPk, accountNumber, name) {
        let mangoAccounts = await this.getMangoAccountsForOwner(group, ownerPk);
        if (mangoAccounts.length === 0) {
            await this.createMangoAccount(group, accountNumber, name);
            mangoAccounts = await this.getMangoAccountsForOwner(group, ownerPk);
        }
        return mangoAccounts[0];
    }
    async createMangoAccount(group, accountNumber, name) {
        return await this.program.methods
            .accountCreate(accountNumber !== null && accountNumber !== void 0 ? accountNumber : 0, 8, 0, 0, 0, name !== null && name !== void 0 ? name : '')
            .accounts({
            group: group.publicKey,
            owner: this.program.provider.wallet.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .rpc();
    }
    async expandMangoAccount(group, account, tokenCount, serum3Count, perpCount, perpOoCount) {
        return await this.program.methods
            .accountExpand(tokenCount, serum3Count, perpCount, perpOoCount)
            .accounts({
            group: group.publicKey,
            account: account.publicKey,
            owner: this.program.provider.wallet.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .rpc();
    }
    async editMangoAccount(group, mangoAccount, name, delegate) {
        return await this.program.methods
            .accountEdit(name !== null && name !== void 0 ? name : null, delegate !== null && delegate !== void 0 ? delegate : null)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
        })
            .rpc({ skipPreflight: true });
    }
    async getMangoAccount(mangoAccount) {
        return mangoAccount_1.MangoAccount.from(mangoAccount.publicKey, await this.program.account.mangoAccount.fetch(mangoAccount.publicKey));
    }
    async getMangoAccountsForOwner(group, ownerPk) {
        return (await this.program.account.mangoAccount.all([
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
            {
                memcmp: {
                    bytes: ownerPk.toBase58(),
                    offset: 40,
                },
            },
        ])).map((pa) => {
            return mangoAccount_1.MangoAccount.from(pa.publicKey, pa.account);
        });
    }
    async closeMangoAccount(group, mangoAccount) {
        return await this.program.methods
            .accountClose()
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            solDestination: mangoAccount.owner,
        })
            .rpc();
    }
    async computeAccountData(group, mangoAccount) {
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [
            mangoAccount,
        ]);
        // Use our custom simulate fn in utils/anchor.ts so signing the tx is not required
        this.program.provider.simulate = anchor_2.simulate;
        let res;
        try {
            res = await this.program.methods
                .computeAccountData()
                .accounts({
                group: group.publicKey,
                account: mangoAccount.publicKey,
            })
                .remainingAccounts(healthRemainingAccounts.map((pk) => ({
                pubkey: pk,
                isWritable: false,
                isSigner: false,
            })))
                .simulate();
        }
        catch (error) {
            console.log(error);
        }
        return mangoAccount_1.MangoAccountData.from(res.events.find((event) => (event.name = 'MangoAccountData')).data);
    }
    async tokenDeposit(group, mangoAccount, tokenName, amount) {
        var _a;
        const bank = group.banksMap.get(tokenName);
        const tokenAccountPk = await (0, utils_1.getAssociatedTokenAddress)(bank.mint, mangoAccount.owner);
        let wrappedSolAccount;
        let preInstructions = [];
        let postInstructions = [];
        const additionalSigners = [];
        if (bank.mint.equals(token_instructions_1.WRAPPED_SOL_MINT)) {
            wrappedSolAccount = new web3_js_1.Keypair();
            const lamports = Math.round(amount * web3_js_1.LAMPORTS_PER_SOL) + 1e7;
            preInstructions = [
                web3_js_1.SystemProgram.createAccount({
                    fromPubkey: mangoAccount.owner,
                    newAccountPubkey: wrappedSolAccount.publicKey,
                    lamports,
                    space: 165,
                    programId: spl_token_1.TOKEN_PROGRAM_ID,
                }),
                (0, token_instructions_1.initializeAccount)({
                    account: wrappedSolAccount.publicKey,
                    mint: token_instructions_1.WRAPPED_SOL_MINT,
                    owner: mangoAccount.owner,
                }),
            ];
            postInstructions = [
                (0, token_instructions_1.closeAccount)({
                    source: wrappedSolAccount.publicKey,
                    destination: mangoAccount.owner,
                    owner: mangoAccount.owner,
                }),
            ];
            additionalSigners.push(wrappedSolAccount);
        }
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], [bank]);
        return await this.program.methods
            .tokenDeposit((0, utils_1.toNativeDecimals)(amount, bank.mintDecimals))
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            bank: bank.publicKey,
            vault: bank.vault,
            tokenAccount: (_a = wrappedSolAccount === null || wrappedSolAccount === void 0 ? void 0 : wrappedSolAccount.publicKey) !== null && _a !== void 0 ? _a : tokenAccountPk,
            tokenAuthority: mangoAccount.owner,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .preInstructions(preInstructions)
            .postInstructions(postInstructions)
            .signers(additionalSigners)
            .rpc({ skipPreflight: true });
    }
    async tokenWithdraw(group, mangoAccount, tokenName, amount, allowBorrow) {
        const bank = group.banksMap.get(tokenName);
        const tokenAccountPk = await (0, utils_1.getAssociatedTokenAddress)(bank.mint, mangoAccount.owner);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], [bank]);
        return await this.program.methods
            .tokenWithdraw((0, utils_1.toNativeDecimals)(amount, bank.mintDecimals), allowBorrow)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            bank: bank.publicKey,
            vault: bank.vault,
            tokenAccount: tokenAccountPk,
            owner: mangoAccount.owner,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .rpc({ skipPreflight: true });
    }
    async tokenWithdrawNative(group, mangoAccount, tokenName, nativeAmount, allowBorrow) {
        const bank = group.banksMap.get(tokenName);
        const tokenAccountPk = await (0, utils_1.getAssociatedTokenAddress)(bank.mint, mangoAccount.owner);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], [bank]);
        return await this.program.methods
            .tokenWithdraw(new anchor_1.BN(nativeAmount), allowBorrow)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            bank: bank.publicKey,
            vault: bank.vault,
            tokenAccount: tokenAccountPk,
            owner: mangoAccount.owner,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .rpc({ skipPreflight: true });
    }
    // Serum
    async serum3RegisterMarket(group, serum3MarketExternalPk, baseBank, quoteBank, marketIndex, name) {
        return await this.program.methods
            .serum3RegisterMarket(marketIndex, name)
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            serumProgram: constants_1.SERUM3_PROGRAM_ID[this.cluster],
            serumMarketExternal: serum3MarketExternalPk,
            baseBank: baseBank.publicKey,
            quoteBank: quoteBank.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .rpc();
    }
    async serum3deregisterMarket(group, serum3MarketName) {
        const serum3Market = group.serum3MarketsMap.get(serum3MarketName);
        return await this.program.methods
            .serum3DeregisterMarket()
            .accounts({
            group: group.publicKey,
            serumMarket: serum3Market.publicKey,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .rpc();
    }
    async serum3GetMarkets(group, baseTokenIndex, quoteTokenIndex) {
        const bumpfbuf = Buffer.alloc(1);
        bumpfbuf.writeUInt8(255);
        const filters = [
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ];
        if (baseTokenIndex) {
            const bbuf = Buffer.alloc(2);
            bbuf.writeUInt16LE(baseTokenIndex);
            filters.push({
                memcmp: {
                    bytes: bs58_1.default.encode(bbuf),
                    offset: 40,
                },
            });
        }
        if (quoteTokenIndex) {
            const qbuf = Buffer.alloc(2);
            qbuf.writeUInt16LE(quoteTokenIndex);
            filters.push({
                memcmp: {
                    bytes: bs58_1.default.encode(qbuf),
                    offset: 42,
                },
            });
        }
        return (await this.program.account.serum3Market.all(filters)).map((tuple) => serum3_1.Serum3Market.from(tuple.publicKey, tuple.account));
    }
    async serum3CreateOpenOrders(group, mangoAccount, marketName) {
        const serum3Market = group.serum3MarketsMap.get(marketName);
        return await this.program.methods
            .serum3CreateOpenOrders()
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            serumMarket: serum3Market.publicKey,
            serumProgram: serum3Market.serumProgram,
            serumMarketExternal: serum3Market.serumMarketExternal,
            owner: this.program.provider.wallet.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .rpc();
    }
    async serum3CloseOpenOrders(group, mangoAccount, serum3MarketName) {
        var _a;
        const serum3Market = group.serum3MarketsMap.get(serum3MarketName);
        const openOrders = (_a = mangoAccount.serum3.find((account) => account.marketIndex === serum3Market.marketIndex)) === null || _a === void 0 ? void 0 : _a.openOrders;
        return await this.program.methods
            .serum3CloseOpenOrders()
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            serumMarket: serum3Market.publicKey,
            serumProgram: serum3Market.serumProgram,
            serumMarketExternal: serum3Market.serumMarketExternal,
            openOrders,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .rpc();
    }
    async serum3PlaceOrder(group, mangoAccount, serum3MarketName, side, price, size, selfTradeBehavior, orderType, clientOrderId, limit) {
        var _a, _b, _c, _d, _e;
        const serum3Market = group.serum3MarketsMap.get(serum3MarketName);
        if (!mangoAccount.findSerum3Account(serum3Market.marketIndex)) {
            await this.serum3CreateOpenOrders(group, mangoAccount, 'BTC/USDC');
            mangoAccount = await this.getMangoAccount(mangoAccount);
        }
        const serum3MarketExternal = group.serum3MarketExternalsMap.get(serum3MarketName);
        const serum3MarketExternalVaultSigner = await web3_js_1.PublicKey.createProgramAddress([
            serum3Market.serumMarketExternal.toBuffer(),
            serum3MarketExternal.decoded.vaultSignerNonce.toArrayLike(Buffer, 'le', 8),
        ], constants_1.SERUM3_PROGRAM_ID[this.cluster]);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [
            mangoAccount,
        ]);
        const limitPrice = serum3MarketExternal.priceNumberToLots(price);
        const maxBaseQuantity = serum3MarketExternal.baseSizeNumberToLots(size);
        const feeTier = (0, serum_1.getFeeTier)(0, 0 /** TODO: fix msrm/srm balance */);
        const rates = (0, serum_1.getFeeRates)(feeTier);
        const maxQuoteQuantity = new anchor_1.BN(serum3MarketExternal.decoded.quoteLotSize.toNumber() *
            (1 + rates.taker) /** TODO: fix taker/maker */).mul(serum3MarketExternal
            .baseSizeNumberToLots(size)
            .mul(serum3MarketExternal.priceNumberToLots(price)));
        return await this.program.methods
            .serum3PlaceOrder(side, limitPrice, maxBaseQuantity, maxQuoteQuantity, selfTradeBehavior, orderType, new anchor_1.BN(clientOrderId), limit)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            openOrders: (_a = mangoAccount.findSerum3Account(serum3Market.marketIndex)) === null || _a === void 0 ? void 0 : _a.openOrders,
            serumMarket: serum3Market.publicKey,
            serumProgram: constants_1.SERUM3_PROGRAM_ID[this.cluster],
            serumMarketExternal: serum3Market.serumMarketExternal,
            marketBids: serum3MarketExternal.bidsAddress,
            marketAsks: serum3MarketExternal.asksAddress,
            marketEventQueue: serum3MarketExternal.decoded.eventQueue,
            marketRequestQueue: serum3MarketExternal.decoded.requestQueue,
            marketBaseVault: serum3MarketExternal.decoded.baseVault,
            marketQuoteVault: serum3MarketExternal.decoded.quoteVault,
            marketVaultSigner: serum3MarketExternalVaultSigner,
            quoteBank: (_b = group.findBank(serum3Market.quoteTokenIndex)) === null || _b === void 0 ? void 0 : _b.publicKey,
            quoteVault: (_c = group.findBank(serum3Market.quoteTokenIndex)) === null || _c === void 0 ? void 0 : _c.vault,
            baseBank: (_d = group.findBank(serum3Market.baseTokenIndex)) === null || _d === void 0 ? void 0 : _d.publicKey,
            baseVault: (_e = group.findBank(serum3Market.baseTokenIndex)) === null || _e === void 0 ? void 0 : _e.vault,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .rpc();
    }
    async serum3CancelAllorders(group, mangoAccount, serum3MarketName, limit) {
        var _a;
        const serum3Market = group.serum3MarketsMap.get(serum3MarketName);
        const serum3MarketExternal = group.serum3MarketExternalsMap.get(serum3MarketName);
        return await this.program.methods
            .serum3CancelAllOrders(limit)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            openOrders: (_a = mangoAccount.findSerum3Account(serum3Market.marketIndex)) === null || _a === void 0 ? void 0 : _a.openOrders,
            serumMarket: serum3Market.publicKey,
            serumProgram: constants_1.SERUM3_PROGRAM_ID[this.cluster],
            serumMarketExternal: serum3Market.serumMarketExternal,
            marketBids: serum3MarketExternal.bidsAddress,
            marketAsks: serum3MarketExternal.asksAddress,
            marketEventQueue: serum3MarketExternal.decoded.eventQueue,
        })
            .rpc();
    }
    async serum3SettleFunds(group, mangoAccount, serum3MarketName) {
        var _a, _b, _c, _d, _e;
        const serum3Market = group.serum3MarketsMap.get(serum3MarketName);
        const serum3MarketExternal = group.serum3MarketExternalsMap.get(serum3MarketName);
        const serum3MarketExternalVaultSigner = 
        // TODO: put into a helper method, and remove copy pasta
        await web3_js_1.PublicKey.createProgramAddress([
            serum3Market.serumMarketExternal.toBuffer(),
            serum3MarketExternal.decoded.vaultSignerNonce.toArrayLike(Buffer, 'le', 8),
        ], constants_1.SERUM3_PROGRAM_ID[this.cluster]);
        return await this.program.methods
            .serum3SettleFunds()
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            openOrders: (_a = mangoAccount.findSerum3Account(serum3Market.marketIndex)) === null || _a === void 0 ? void 0 : _a.openOrders,
            serumMarket: serum3Market.publicKey,
            serumProgram: constants_1.SERUM3_PROGRAM_ID[this.cluster],
            serumMarketExternal: serum3Market.serumMarketExternal,
            marketBaseVault: serum3MarketExternal.decoded.baseVault,
            marketQuoteVault: serum3MarketExternal.decoded.quoteVault,
            marketVaultSigner: serum3MarketExternalVaultSigner,
            quoteBank: (_b = group.findBank(serum3Market.quoteTokenIndex)) === null || _b === void 0 ? void 0 : _b.publicKey,
            quoteVault: (_c = group.findBank(serum3Market.quoteTokenIndex)) === null || _c === void 0 ? void 0 : _c.vault,
            baseBank: (_d = group.findBank(serum3Market.baseTokenIndex)) === null || _d === void 0 ? void 0 : _d.publicKey,
            baseVault: (_e = group.findBank(serum3Market.baseTokenIndex)) === null || _e === void 0 ? void 0 : _e.vault,
        })
            .rpc();
    }
    async serum3CancelOrder(group, mangoAccount, serum3MarketName, side, orderId) {
        var _a;
        const serum3Market = group.serum3MarketsMap.get(serum3MarketName);
        const serum3MarketExternal = group.serum3MarketExternalsMap.get(serum3MarketName);
        return await this.program.methods
            .serum3CancelOrder(side, orderId)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            openOrders: (_a = mangoAccount.findSerum3Account(serum3Market.marketIndex)) === null || _a === void 0 ? void 0 : _a.openOrders,
            serumMarket: serum3Market.publicKey,
            serumProgram: constants_1.SERUM3_PROGRAM_ID[this.cluster],
            serumMarketExternal: serum3Market.serumMarketExternal,
            marketBids: serum3MarketExternal.bidsAddress,
            marketAsks: serum3MarketExternal.asksAddress,
            marketEventQueue: serum3MarketExternal.decoded.eventQueue,
        })
            .rpc();
    }
    async getSerum3Orders(group, serum3MarketName) {
        const serum3MarketExternal = group.serum3MarketExternalsMap.get(serum3MarketName);
        // TODO: filter for mango account
        return await serum3MarketExternal.loadOrdersForOwner(this.program.provider.connection, group.publicKey);
    }
    /// perps
    async perpCreateMarket(group, oraclePk, perpMarketIndex, name, oracleConfFilter, baseTokenIndex, baseTokenDecimals, quoteTokenIndex, quoteLotSize, baseLotSize, maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee, makerFee, takerFee, minFunding, maxFunding, impactQuantity) {
        const bids = new web3_js_1.Keypair();
        const asks = new web3_js_1.Keypair();
        const eventQueue = new web3_js_1.Keypair();
        return await this.program.methods
            .perpCreateMarket(perpMarketIndex, name, {
            confFilter: {
                val: I80F48_1.I80F48.fromNumber(oracleConfFilter).getData(),
            },
        }, // future: nested custom types dont typecheck, fix if possible?
        baseTokenIndex, baseTokenDecimals, new anchor_1.BN(quoteLotSize), new anchor_1.BN(baseLotSize), maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee, makerFee, takerFee, minFunding, maxFunding, new anchor_1.BN(impactQuantity))
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            oracle: oraclePk,
            bids: bids.publicKey,
            asks: asks.publicKey,
            eventQueue: eventQueue.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .preInstructions([
            // TODO: try to pick up sizes of bookside and eventqueue from IDL, so we can stay in sync with program
            // book sides
            web3_js_1.SystemProgram.createAccount({
                programId: this.program.programId,
                space: 8 + 98584,
                lamports: await this.program.provider.connection.getMinimumBalanceForRentExemption(8 + 98584),
                fromPubkey: this.program.provider.wallet
                    .publicKey,
                newAccountPubkey: bids.publicKey,
            }),
            web3_js_1.SystemProgram.createAccount({
                programId: this.program.programId,
                space: 8 + 98584,
                lamports: await this.program.provider.connection.getMinimumBalanceForRentExemption(8 + 98584),
                fromPubkey: this.program.provider.wallet
                    .publicKey,
                newAccountPubkey: asks.publicKey,
            }),
            // event queue
            web3_js_1.SystemProgram.createAccount({
                programId: this.program.programId,
                space: 8 + 4 * 2 + 8 + 488 * 208,
                lamports: await this.program.provider.connection.getMinimumBalanceForRentExemption(8 + 4 * 2 + 8 + 488 * 208),
                fromPubkey: this.program.provider.wallet
                    .publicKey,
                newAccountPubkey: eventQueue.publicKey,
            }),
        ])
            .signers([bids, asks, eventQueue])
            .rpc();
    }
    async perpEditMarket(group, perpMarketName, oracle, oracleConfFilter, baseTokenIndex, baseTokenDecimals, maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee, makerFee, takerFee, minFunding, maxFunding, impactQuantity) {
        const perpMarket = group.perpMarketsMap.get(perpMarketName);
        return await this.program.methods
            .perpEditMarket(oracle, {
            confFilter: {
                val: I80F48_1.I80F48.fromNumber(oracleConfFilter).getData(),
            },
        }, // future: nested custom types dont typecheck, fix if possible?
        baseTokenIndex, baseTokenDecimals, maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee, makerFee, takerFee, minFunding, maxFunding, new anchor_1.BN(impactQuantity))
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            perpMarket: perpMarket.publicKey,
        })
            .rpc();
    }
    async perpCloseMarket(group, perpMarketName) {
        const perpMarket = group.perpMarketsMap.get(perpMarketName);
        return await this.program.methods
            .perpCloseMarket()
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            perpMarket: perpMarket.publicKey,
            asks: perpMarket.asks,
            bids: perpMarket.bids,
            eventQueue: perpMarket.eventQueue,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .rpc();
    }
    async perpGetMarkets(group, baseTokenIndex) {
        const bumpfbuf = Buffer.alloc(1);
        bumpfbuf.writeUInt8(255);
        const filters = [
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ];
        if (baseTokenIndex) {
            const bbuf = Buffer.alloc(2);
            bbuf.writeUInt16LE(baseTokenIndex);
            filters.push({
                memcmp: {
                    bytes: bs58_1.default.encode(bbuf),
                    offset: 40,
                },
            });
        }
        return (await this.program.account.perpMarket.all(filters)).map((tuple) => perp_1.PerpMarket.from(tuple.publicKey, tuple.account));
    }
    async perpPlaceOrder(group, mangoAccount, perpMarketName, side, price, quantity, maxQuoteQuantity, clientOrderId, orderType, expiryTimestamp, limit) {
        const perpMarket = group.perpMarketsMap.get(perpMarketName);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [
            mangoAccount,
        ]);
        const [nativePrice, nativeQuantity] = perpMarket.uiToNativePriceQuantity(price, quantity);
        const maxQuoteQuantityLots = maxQuoteQuantity
            ? perpMarket.uiQuoteToLots(maxQuoteQuantity)
            : utils_1.I64_MAX_BN;
        await this.program.methods
            .perpPlaceOrder(side, nativePrice, nativeQuantity, maxQuoteQuantityLots, new anchor_1.BN(clientOrderId), orderType, new anchor_1.BN(expiryTimestamp), limit)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            perpMarket: perpMarket.publicKey,
            asks: perpMarket.asks,
            bids: perpMarket.bids,
            eventQueue: perpMarket.eventQueue,
            oracle: perpMarket.oracle,
            owner: this.program.provider.wallet.publicKey,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .rpc();
    }
    async marginTrade({ group, mangoAccount, inputToken, amountIn, outputToken, userDefinedInstructions, }) {
        const inputBank = group.banksMap.get(inputToken);
        const outputBank = group.banksMap.get(outputToken);
        if (!inputBank || !outputBank)
            throw new Error('Invalid token');
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], [inputBank, outputBank]);
        const parsedHealthAccounts = healthRemainingAccounts.map((pk) => ({
            pubkey: pk,
            isWritable: false,
            isSigner: false,
        }));
        console.log('1');
        /*
         * Find or create associated token accounts
         */
        const inputTokenAccountPk = await (0, utils_1.getAssociatedTokenAddress)(inputBank.mint, mangoAccount.owner);
        const inputTokenAccExists = await this.program.provider.connection.getAccountInfo(inputTokenAccountPk);
        const preInstructions = [];
        if (!inputTokenAccExists) {
            preInstructions.push(spl_token_1.Token.createAssociatedTokenAccountInstruction(mangoAccount.owner, inputTokenAccountPk, mangoAccount.owner, inputBank.mint, spl_token_1.TOKEN_PROGRAM_ID, spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID));
        }
        const outputTokenAccountPk = await (0, utils_1.getAssociatedTokenAddress)(outputBank.mint, mangoAccount.owner);
        const outputTokenAccExists = await this.program.provider.connection.getAccountInfo(outputTokenAccountPk);
        if (!outputTokenAccExists) {
            preInstructions.push(spl_token_1.Token.createAssociatedTokenAccountInstruction(mangoAccount.owner, outputTokenAccountPk, mangoAccount.owner, outputBank.mint, spl_token_1.TOKEN_PROGRAM_ID, spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID));
        }
        console.log('2');
        if (preInstructions.length) {
            const tx = new web3_js_1.Transaction();
            for (const ix of preInstructions) {
                tx.add(ix);
            }
            console.log('preInstructions', preInstructions);
            await this.program.provider.sendAndConfirm(tx);
        }
        console.log('3');
        const inputBankAccount = {
            pubkey: inputBank.publicKey,
            isWritable: true,
            isSigner: false,
        };
        const outputBankAccount = {
            pubkey: outputBank.publicKey,
            isWritable: true,
            isSigner: false,
        };
        const inputBankVault = {
            pubkey: inputBank.vault,
            isWritable: true,
            isSigner: false,
        };
        const outputBankVault = {
            pubkey: outputBank.vault,
            isWritable: true,
            isSigner: false,
        };
        const inputATA = {
            pubkey: inputTokenAccountPk,
            isWritable: true,
            isSigner: false,
        };
        const outputATA = {
            pubkey: outputTokenAccountPk,
            isWritable: false,
            isSigner: false,
        };
        const flashLoanEndIx = await this.program.methods
            .flashLoanEnd()
            .accounts({
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
        })
            .remainingAccounts([
            ...parsedHealthAccounts,
            inputBankVault,
            outputBankVault,
            inputATA,
            {
                isWritable: true,
                pubkey: outputTokenAccountPk,
                isSigner: false,
            },
        ])
            .instruction();
        console.log('4');
        // userDefinedInstructions.push(flashLoanEndIx);
        const flashLoanBeginIx = await this.program.methods
            .flashLoanBegin([
            (0, utils_1.toNativeDecimals)(amountIn, inputBank.mintDecimals),
            new anchor_1.BN(0) /* we don't care about borrowing the target amount, this is just a dummy */,
        ])
            .accounts({
            group: group.publicKey,
            instructions: web3_js_1.SYSVAR_INSTRUCTIONS_PUBKEY,
        })
            .remainingAccounts([
            inputBankAccount,
            outputBankAccount,
            inputBankVault,
            outputBankVault,
            inputATA,
            outputATA,
        ])
            .instruction();
        const tx = new web3_js_1.Transaction();
        tx.add(flashLoanBeginIx);
        for (const i of userDefinedInstructions) {
            tx.add(i);
        }
        tx.add(flashLoanEndIx);
        return this.program.provider.sendAndConfirm(tx);
    }
    async updateIndexAndRate(group, tokenName) {
        const bank = group.banksMap.get(tokenName);
        const mintInfo = group.mintInfosMap.get(bank.tokenIndex);
        await this.program.methods
            .tokenUpdateIndexAndRate()
            .accounts({
            group: group.publicKey,
            mintInfo: mintInfo.publicKey,
            oracle: mintInfo.oracle,
            instructions: web3_js_1.SYSVAR_INSTRUCTIONS_PUBKEY,
        })
            .remainingAccounts([
            {
                pubkey: bank.publicKey,
                isWritable: true,
                isSigner: false,
            },
        ])
            .rpc();
    }
    /// liquidations
    async liqTokenWithToken(group, liqor, liqee, assetTokenName, liabTokenName, maxLiabTransfer) {
        const assetBank = group.banksMap.get(assetTokenName);
        const liabBank = group.banksMap.get(liabTokenName);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Scanning, group, [liqor, liqee], [assetBank, liabBank]);
        const parsedHealthAccounts = healthRemainingAccounts.map((pk) => ({
            pubkey: pk,
            isWritable: pk.equals(assetBank.publicKey) || pk.equals(liabBank.publicKey)
                ? true
                : false,
            isSigner: false,
        }));
        await this.program.methods
            .liqTokenWithToken(assetBank.tokenIndex, liabBank.tokenIndex, {
            val: I80F48_1.I80F48.fromNumber(maxLiabTransfer).getData(),
        })
            .accounts({
            group: group.publicKey,
            liqor: liqor.publicKey,
            liqee: liqee.publicKey,
            liqorOwner: liqor.owner,
        })
            .remainingAccounts(parsedHealthAccounts)
            .rpc();
    }
    /// static
    static connect(provider, cluster, programId) {
        // TODO: use IDL on chain or in repository? decide...
        // Alternatively we could fetch IDL from chain.
        // const idl = await Program.fetchIdl(MANGO_V4_ID, provider);
        const idl = mango_v4_1.IDL;
        return new MangoClient(new anchor_1.Program(idl, programId, provider), programId, cluster);
    }
    static connectForGroupName(provider, groupName) {
        // TODO: use IDL on chain or in repository? decide...
        // Alternatively we could fetch IDL from chain.
        // const idl = await Program.fetchIdl(MANGO_V4_ID, provider);
        const idl = mango_v4_1.IDL;
        const id = ids_1.Id.fromIds(groupName);
        return new MangoClient(new anchor_1.Program(idl, new web3_js_1.PublicKey(id.mangoProgramId), provider), new web3_js_1.PublicKey(id.mangoProgramId), id.cluster, groupName);
    }
    /// private
    buildHealthRemainingAccounts(retriever, group, mangoAccounts, banks /** TODO for serum3PlaceOrder we are just ingoring this atm */) {
        if (retriever === AccountRetriever.Fixed) {
            return this.buildFixedAccountRetrieverHealthAccounts(group, mangoAccounts[0], banks);
        }
        else {
            return this.buildScanningAccountRetrieverHealthAccounts(group, mangoAccounts, banks);
        }
    }
    buildFixedAccountRetrieverHealthAccounts(group, mangoAccount, banks /** TODO for serum3PlaceOrder we are just ingoring this atm */) {
        const healthRemainingAccounts = [];
        const tokenIndices = mangoAccount.tokens
            .filter((token) => token.tokenIndex !== 65535)
            .map((token) => token.tokenIndex);
        if (banks === null || banks === void 0 ? void 0 : banks.length) {
            for (const bank of banks) {
                tokenIndices.push(bank.tokenIndex);
            }
        }
        const mintInfos = [...new Set(tokenIndices)].map((tokenIndex) => group.mintInfosMap.get(tokenIndex));
        healthRemainingAccounts.push(...mintInfos.map((mintInfo) => mintInfo.firstBank()));
        healthRemainingAccounts.push(...mintInfos.map((mintInfo) => mintInfo.oracle));
        healthRemainingAccounts.push(...mangoAccount.serum3
            .filter((serum3Account) => serum3Account.marketIndex !== 65535)
            .map((serum3Account) => serum3Account.openOrders));
        healthRemainingAccounts.push(...mangoAccount.perps
            .filter((perp) => perp.marketIndex !== 65535)
            .map((perp) => Array.from(group.perpMarketsMap.values()).filter((perpMarket) => perpMarket.perpMarketIndex === perp.marketIndex)[0].publicKey));
        return healthRemainingAccounts;
    }
    buildScanningAccountRetrieverHealthAccounts(group, mangoAccounts, banks /** TODO for serum3PlaceOrder we are just ingoring this atm */) {
        const healthRemainingAccounts = [];
        let tokenIndices = [];
        for (const mangoAccount of mangoAccounts) {
            tokenIndices.push(...mangoAccount.tokens
                .filter((token) => token.tokenIndex !== 65535)
                .map((token) => token.tokenIndex));
        }
        tokenIndices = [...new Set(tokenIndices)];
        if (banks === null || banks === void 0 ? void 0 : banks.length) {
            for (const bank of banks) {
                tokenIndices.push(bank.tokenIndex);
            }
        }
        const mintInfos = [...new Set(tokenIndices)].map((tokenIndex) => group.mintInfosMap.get(tokenIndex));
        healthRemainingAccounts.push(...mintInfos.map((mintInfo) => mintInfo.firstBank()));
        healthRemainingAccounts.push(...mintInfos.map((mintInfo) => mintInfo.oracle));
        for (const mangoAccount of mangoAccounts) {
            healthRemainingAccounts.push(...mangoAccount.serum3
                .filter((serum3Account) => serum3Account.marketIndex !== 65535)
                .map((serum3Account) => serum3Account.openOrders));
        }
        for (const mangoAccount of mangoAccounts) {
            healthRemainingAccounts.push(...mangoAccount.perps
                .filter((perp) => perp.marketIndex !== 65535)
                .map((perp) => Array.from(group.perpMarketsMap.values()).filter((perpMarket) => perpMarket.perpMarketIndex === perp.marketIndex)[0].publicKey));
        }
        return healthRemainingAccounts;
    }
}
exports.MangoClient = MangoClient;
//# sourceMappingURL=client.js.map