import { BorshAccountsCoder } from '@project-serum/anchor';
import { Market } from '@project-serum/serum';
import { parsePriceData } from '@pythnetwork/client';
import { SERUM3_PROGRAM_ID } from '../constants';
import { Id } from '../ids';
import { Bank, MintInfo } from './bank';
import { I80F48, ONE_I80F48 } from './I80F48';
import { PerpMarket } from './perp';
import { Serum3Market } from './serum3';
export class Group {
    constructor(publicKey, creator, groupNum, admin, fastListingAdmin, insuranceMint, insuranceVault, testing, version, banksMap, serum3MarketsMap, serum3MarketExternalsMap, perpMarketsMap, mintInfosMap, oraclesMap) {
        this.publicKey = publicKey;
        this.creator = creator;
        this.groupNum = groupNum;
        this.admin = admin;
        this.fastListingAdmin = fastListingAdmin;
        this.insuranceMint = insuranceMint;
        this.insuranceVault = insuranceVault;
        this.testing = testing;
        this.version = version;
        this.banksMap = banksMap;
        this.serum3MarketsMap = serum3MarketsMap;
        this.serum3MarketExternalsMap = serum3MarketExternalsMap;
        this.perpMarketsMap = perpMarketsMap;
        this.mintInfosMap = mintInfosMap;
        this.oraclesMap = oraclesMap;
    }
    static from(publicKey, obj) {
        return new Group(publicKey, obj.creator, obj.groupNum, obj.admin, obj.fastListingAdmin, obj.insuranceMint, obj.insuranceVault, obj.testing, obj.version, new Map(), new Map(), new Map(), new Map(), new Map(), new Map());
    }
    findBank(tokenIndex) {
        return Array.from(this.banksMap.values()).find((bank) => bank.tokenIndex === tokenIndex);
    }
    findSerum3Market(marketIndex) {
        return Array.from(this.serum3MarketsMap.values()).find((serum3Market) => serum3Market.marketIndex === marketIndex);
    }
    async reloadAll(client) {
        let ids = undefined;
        if (client.groupName) {
            ids = Id.fromIds(client.groupName);
        }
        // console.time('group.reload');
        await Promise.all([
            this.reloadBanks(client, ids),
            this.reloadMintInfos(client, ids),
            this.reloadSerum3Markets(client, ids),
            this.reloadPerpMarkets(client, ids),
        ]);
        await Promise.all([
            // requires reloadBanks to have finished loading
            this.reloadBankPrices(client, ids),
            // requires reloadSerum3Markets to have finished loading
            this.reloadSerum3ExternalMarkets(client, ids),
        ]);
        // console.timeEnd('group.reload');
    }
    async reloadBanks(client, ids) {
        let banks;
        if (ids) {
            banks = (await client.program.account.bank.fetchMultiple(ids.getBanks())).map((account, index) => Bank.from(ids.getBanks()[index], account));
        }
        else {
            banks = await client.getBanksForGroup(this);
        }
        this.banksMap = new Map(banks.map((bank) => [bank.name, bank]));
    }
    async reloadMintInfos(client, ids) {
        let mintInfos;
        if (ids) {
            mintInfos = (await client.program.account.mintInfo.fetchMultiple(ids.getMintInfos())).map((account, index) => MintInfo.from(ids.getMintInfos()[index], account));
        }
        else {
            mintInfos = await client.getMintInfosForGroup(this);
        }
        this.mintInfosMap = new Map(mintInfos.map((mintInfo) => {
            return [mintInfo.tokenIndex, mintInfo];
        }));
    }
    async reloadSerum3Markets(client, ids) {
        let serum3Markets;
        if (ids) {
            serum3Markets = (await client.program.account.serum3Market.fetchMultiple(ids.getSerum3Markets())).map((account, index) => Serum3Market.from(ids.getSerum3Markets()[index], account));
        }
        else {
            serum3Markets = await client.serum3GetMarkets(this);
        }
        this.serum3MarketsMap = new Map(serum3Markets.map((serum3Market) => [serum3Market.name, serum3Market]));
    }
    async reloadSerum3ExternalMarkets(client, ids) {
        const externalMarkets = await Promise.all(Array.from(this.serum3MarketsMap.values()).map((serum3Market) => Market.load(client.program.provider.connection, serum3Market.serumMarketExternal, { commitment: client.program.provider.connection.commitment }, SERUM3_PROGRAM_ID[client.cluster])));
        this.serum3MarketExternalsMap = new Map(Array.from(this.serum3MarketsMap.values()).map((serum3Market, index) => [
            serum3Market.name,
            externalMarkets[index],
        ]));
    }
    async reloadPerpMarkets(client, ids) {
        let perpMarkets;
        if (ids) {
            perpMarkets = (await client.program.account.perpMarket.fetchMultiple(ids.getPerpMarkets())).map((account, index) => PerpMarket.from(ids.getPerpMarkets()[index], account));
        }
        else {
            perpMarkets = await client.perpGetMarkets(this);
        }
        this.perpMarketsMap = new Map(perpMarkets.map((perpMarket) => [perpMarket.name, perpMarket]));
    }
    async reloadBankPrices(client, ids) {
        const banks = Array.from(this === null || this === void 0 ? void 0 : this.banksMap, ([, value]) => value);
        const oracles = banks.map((b) => b.oracle);
        const prices = await client.program.provider.connection.getMultipleAccountsInfo(oracles);
        const coder = new BorshAccountsCoder(client.program.idl);
        for (const [index, price] of prices.entries()) {
            if (banks[index].name === 'USDC') {
                banks[index].price = ONE_I80F48;
            }
            else {
                // TODO: Implement switchboard oracle type
                if (!BorshAccountsCoder.accountDiscriminator('stubOracle').compare(price.data.slice(0, 8))) {
                    const stubOracle = coder.decode('stubOracle', price.data);
                    banks[index].price = new I80F48(stubOracle.price.val);
                }
                else {
                    banks[index].price = I80F48.fromNumber(parsePriceData(price.data).previousPrice);
                }
            }
        }
    }
    toString() {
        let res = 'Group\n';
        res = res + ' pk: ' + this.publicKey.toString();
        res =
            res +
                '\n mintInfos:' +
                Array.from(this.mintInfosMap.entries())
                    .map((mintInfoTuple) => '  \n' + mintInfoTuple[0] + ') ' + mintInfoTuple[1].toString())
                    .join(', ');
        return res;
    }
}
//# sourceMappingURL=group.js.map