/// <reference types="bn.js" />
import { BN } from '@project-serum/anchor';
import { PublicKey } from '@solana/web3.js';
import { MangoClient } from '../client';
import { Bank } from './bank';
import { Group } from './group';
import { HealthCache, HealthCacheDto } from './healthCache';
import { I80F48, I80F48Dto } from './I80F48';
export declare class MangoAccount {
    publicKey: PublicKey;
    group: PublicKey;
    owner: PublicKey;
    delegate: PublicKey;
    accountNum: number;
    accountData: MangoAccountData;
    tokens: TokenPosition[];
    serum3: Serum3Orders[];
    perps: PerpPositions[];
    name: string;
    static from(publicKey: PublicKey, obj: {
        group: PublicKey;
        owner: PublicKey;
        name: number[];
        delegate: PublicKey;
        beingLiquidated: number;
        isBankrupt: number;
        accountNum: number;
        bump: number;
        netDeposits: number;
        netSettled: number;
        headerVersion: number;
        tokens: unknown;
        serum3: unknown;
        perps: unknown;
        perpOpenOrders: unknown;
    }): MangoAccount;
    constructor(publicKey: PublicKey, group: PublicKey, owner: PublicKey, name: number[], delegate: PublicKey, beingLiquidated: number, isBankrupt: number, accountNum: number, bump: number, netDeposits: number, netSettled: number, headerVersion: number, tokens: TokenPositionDto[], serum3: Serum3PositionDto[], perps: PerpPositionDto[], perpOpenOrders: PerpPositionDto[], accountData: MangoAccountData);
    reload(client: MangoClient, group: Group): Promise<void>;
    reloadAccountData(client: MangoClient, group: Group): Promise<void>;
    findToken(tokenIndex: number): TokenPosition | undefined;
    findSerum3Account(marketIndex: number): Serum3Orders | undefined;
    getNative(bank: Bank): I80F48;
    static getEquivalentNativeUsdcPosition(sourceBank: Bank, nativeTokenPosition: TokenPosition): I80F48;
    static getEquivalentNativeTokenPosition(targetBank: Bank, nativeUsdcPosition: I80F48): I80F48;
    getNativeDeposits(bank: Bank): I80F48;
    getNativeBorrows(bank: Bank): I80F48;
    getUi(bank: Bank): number;
    deposits(bank: Bank): number;
    borrows(bank: Bank): number;
    getHealth(healthType: HealthType): I80F48;
    /**
     * TODO: this is incorrect, getAssetsVal and getLiabsVal are in equity, and not in given health type.
     * Wait for dev to be deployed to mainnet, and then we can adapt this.
     */
    getHealthRatio(healthType: HealthType): I80F48;
    /**
     * Sum of all the assets i.e. token deposits, borrows, total assets in spot open orders, (perps positions is todo) in terms of quote value.
     */
    getEquity(): I80F48;
    /**
     * The amount of native quote you could withdraw against your existing assets.
     */
    getCollateralValue(): I80F48;
    /**
     * Similar to getEquity, but only the sum of all positive assets.
     */
    getAssetsVal(): I80F48;
    /**
     * Similar to getEquity, but only the sum of all negative assets. Note: return value would be negative.
     */
    getLiabsVal(): I80F48;
    /**
     * The amount of given native token you can borrow, considering all existing assets as collateral except the deposits for this token.
     * Note 1: The existing native deposits need to be added to get the full amount that could be withdrawn.
     * Note 2: The group might have less native deposits than what this returns. TODO: loan origination fees
     */
    getMaxWithdrawWithBorrowForToken(group: Group, tokenName: string): I80F48;
    /**
     * The amount of given source native token you can swap to a target token considering all existing assets as collateral.
     * note: slippageAndFeesFactor is a normalized number, <1, e.g. a slippage of 5% and some fees which are 1%, then slippageAndFeesFactor = 0.94
     * the factor is used to compute how much target can be obtained by swapping source TODO: loan origination fees
     */
    getMaxSourceForTokenSwap(group: Group, sourceTokenName: string, targetTokenName: string, slippageAndFeesFactor: number): I80F48;
    /**
     * Simulates new health after applying tokenChanges to the token positions. Useful to simulate health after a potential swap.
     */
    simHealthWithTokenPositionChanges(group: Group, tokenChanges: {
        tokenName: string;
        tokenAmount: number;
    }[]): I80F48;
    /**
     * The remaining native quote margin available for given market.
     *
     * TODO: this is a very bad estimation atm.
     * It assumes quote asset is always USDC,
     * it assumes that there are no interaction effects,
     * it assumes that there are no existing borrows for either of the tokens in the market.
     */
    getSerum3MarketMarginAvailable(group: Group, marketName: string): I80F48;
    /**
     * The remaining native quote margin available for given market.
     *
     * TODO: this is a very bad estimation atm.
     * It assumes quote asset is always USDC,
     * it assumes that there are no interaction effects,
     * it assumes that there are no existing borrows for either of the tokens in the market.
     */
    getPerpMarketMarginAvailable(group: Group, marketName: string): I80F48;
    tokensActive(): TokenPosition[];
    serum3Active(): Serum3Orders[];
    perpActive(): PerpPositions[];
    toString(group?: Group): string;
}
export declare class TokenPosition {
    indexedPosition: I80F48;
    tokenIndex: number;
    inUseCount: number;
    static TokenIndexUnset: number;
    static from(dto: TokenPositionDto): TokenPosition;
    constructor(indexedPosition: I80F48, tokenIndex: number, inUseCount: number);
    isActive(): boolean;
    native(bank: Bank): I80F48;
    ui(bank: Bank): number;
    uiDeposits(bank: Bank): number;
    uiBorrows(bank: Bank): number;
    toString(group?: Group): string;
}
export declare class TokenPositionDto {
    indexedPosition: I80F48Dto;
    tokenIndex: number;
    inUseCount: number;
    reserved: number[];
    constructor(indexedPosition: I80F48Dto, tokenIndex: number, inUseCount: number, reserved: number[]);
}
export declare class Serum3Orders {
    openOrders: PublicKey;
    marketIndex: number;
    baseTokenIndex: number;
    quoteTokenIndex: number;
    static Serum3MarketIndexUnset: number;
    static from(dto: Serum3PositionDto): Serum3Orders;
    constructor(openOrders: PublicKey, marketIndex: number, baseTokenIndex: number, quoteTokenIndex: number);
    isActive(): boolean;
}
export declare class Serum3PositionDto {
    openOrders: PublicKey;
    marketIndex: number;
    baseTokenIndex: number;
    quoteTokenIndex: number;
    reserved: number[];
    constructor(openOrders: PublicKey, marketIndex: number, baseTokenIndex: number, quoteTokenIndex: number, reserved: number[]);
}
export declare class PerpPositions {
    marketIndex: number;
    basePositionLots: number;
    quotePositionNative: number;
    bidsBaseLots: number;
    asksBaseLots: number;
    takerBaseLots: number;
    takerQuoteLots: number;
    static PerpMarketIndexUnset: number;
    static from(dto: PerpPositionDto): PerpPositions;
    constructor(marketIndex: number, basePositionLots: number, quotePositionNative: number, bidsBaseLots: number, asksBaseLots: number, takerBaseLots: number, takerQuoteLots: number);
    isActive(): boolean;
}
export declare class PerpPositionDto {
    marketIndex: number;
    reserved: [];
    basePositionLots: BN;
    quotePositionNative: {
        val: BN;
    };
    bidsBaseLots: BN;
    asksBaseLots: BN;
    takerBaseLots: BN;
    takerQuoteLots: BN;
    constructor(marketIndex: number, reserved: [], basePositionLots: BN, quotePositionNative: {
        val: BN;
    }, bidsBaseLots: BN, asksBaseLots: BN, takerBaseLots: BN, takerQuoteLots: BN);
}
export declare class HealthType {
    static maint: {
        maint: {};
    };
    static init: {
        init: {};
    };
}
export declare class MangoAccountData {
    healthCache: HealthCache;
    initHealth: I80F48;
    maintHealth: I80F48;
    equity: Equity;
    constructor(healthCache: HealthCache, initHealth: I80F48, maintHealth: I80F48, equity: Equity);
    static from(event: {
        healthCache: HealthCacheDto;
        initHealth: I80F48Dto;
        maintHealth: I80F48Dto;
        equity: {
            tokens: [{
                tokenIndex: number;
                value: I80F48Dto;
            }];
            perps: [{
                perpMarketIndex: number;
                value: I80F48Dto;
            }];
        };
        initHealthLiabs: I80F48Dto;
        tokenAssets: any;
    }): MangoAccountData;
}
export declare class Equity {
    tokens: TokenEquity[];
    perps: PerpEquity[];
    constructor(tokens: TokenEquity[], perps: PerpEquity[]);
    static from(dto: EquityDto): Equity;
}
export declare class TokenEquity {
    tokenIndex: number;
    value: I80F48;
    constructor(tokenIndex: number, value: I80F48);
}
export declare class PerpEquity {
    perpMarketIndex: number;
    value: I80F48;
    constructor(perpMarketIndex: number, value: I80F48);
}
export declare class EquityDto {
    tokens: {
        tokenIndex: number;
        value: I80F48Dto;
    }[];
    perps: {
        perpMarketIndex: number;
        value: I80F48Dto;
    }[];
}
